

#include "config.h"
#include "gfortran.h"
#include <ctype.h>

//#define DISABLE_STATEMENTS

#define PDB_ENABLE_TYPES
//#define PDB_ENABLE_RCALL

#define PDB_ACCESS_PRIVATE 0
#define PDB_ACCESS_PUBLIC 1

#define PDB_TYPE_FLOAT 0
#define PDB_TYPE_INTEGER 1

typedef enum {
  KIND_NA, KIND_VOID, KIND_INT, KIND_LOGICAL, KIND_FLOAT, KIND_FUNC, KIND_CHAR, 
  KIND_ARRAY, KIND_COMPLEX, KIND_UNSPECFUNC, KIND_BLDAT, KIND_MOD, KIND_POINTER, KIND_DERIVED
} 
ykind;
const char *ykind_strings[14] = {"NA", "fvoid", "fint", "flogic", "ffloat", "ffunc", "fchar", "farray",
				 "fcmplx", "funspecfunc", "fbldat", "fmod", "fptr", "fderived"};


typedef enum {
  IKIND_NA, IKIND_CHAR, IKIND_SCHAR, IKIND_UCHAR, IKIND_WCHAR, IKIND_SHORT, IKIND_USHORT,
  IKIND_INT, IKIND_UINT, IKIND_ULONG, IKIND_LONGLONG, IKIND_ULONGLONG
} yikind;
const char *yikind_strings[13] = {"NA", "char", "schar", "uchar", "wchar", "short", "ushort", "int",
				 "uint", "long", "ulong", "longlong", "ulonglong" };

typedef enum {
  FKIND_NA, FKIND_FLOAT, FKIND_DOUBLE, FKIND_LONGDOUBLE
} yfkind;
const char *yfkind_strings[4] = {"NA", "float", "dbl", "longdbl" };


const char *yshape[6] = {"NA", "explicit", "asmdshape", "deferred", "asmdsize", "NA" };

const char *access_strings[3] = {"NA", "pub", "priv" };



typedef struct pdbType {
  int id;
  const char *name;
  char idString[128];
  ykind kind;
  yikind ikind;
  yfkind fkind;

  struct pdbType *returnType;
  struct pdbArg *args;
  struct pdbGroupMember *members;

  gfc_access access;

  /* for arrays */
  struct pdbType *baseType; 
  gfc_array_spec as;

  struct pdbType *nextLink;
} pdbType;
pdbType *pdbFirstType = NULL;
pdbType *pdbLastType = NULL;




typedef struct derivedListNode {
  gfc_symbol *sym;
  struct pdbType *type;
  struct derviedListNode *next;
} derivedListNode;
struct derivedListNode *derivedFirst = NULL;
struct derivedListNode *derivedLast = NULL;


typedef struct pdbArg {
  struct pdbType *type;
  const char *name;
  struct pdbArg *nextLink;
} pdbArg;




typedef struct pdbLoc {
  int file;
  int line;
  int column;
} pdbLoc;

typedef struct pdbStatement{
  int id, kind;
  pdbLoc start;
  pdbLoc end;
  
  struct pdbStatement *nextstmt;
  struct pdbStatement *downstmt;
  struct pdbStatement *extrastmt;
  struct pdbStatement *nextLink;

  gfc_st_label *label;
  gfc_code *code;

} pdbStatement;


typedef struct pdbRoutine {
  int id;
  int appearsInThisFile;
  const char *name;
  gfc_symbol *sym;
  struct pdbType *type;
  pdbLoc location;
  pdbLoc rpos_start;
  pdbLoc rpos_end;
  struct pdbStatement *statements;
  struct pdbRcall *rcalls;
  struct pdbRoutine *nextLink;
} pdbRoutine;
pdbRoutine *pdbFirstRoutine = NULL;
pdbRoutine *pdbLastRoutine = NULL;

typedef struct pdbRcall {
  struct gfc_expr *expr;
  struct pdbRoutine *routine;
  pdbLoc location;
  int routineID;
  struct pdbRcall *nextLink;
} pdbRcall;


pdbRcall *pdbFirstRcall = NULL;
pdbRcall *pdbLastRcall = NULL;

pdbRcall *createRcall(gfc_expr *expr);
pdbRcall *createRcallS(gfc_code *code);

static pdbRoutine *pdbGetRoutine(gfc_symbol *sym);


int pdbStartColumn, pdbEndColumn;


pdbRcall *createRcall(gfc_expr *expr) {
  pdbRcall *rcall;

  rcall = (pdbRcall *) gfc_getmem (sizeof(pdbRcall));
  rcall->expr = expr;


  rcall->routine = NULL;
  rcall->routineID = -1;
/*   rcall->routine = NULL; /\* we have to look it up later *\/ */
  if (expr->symtree) {
/*     printf ("> looking up '%s'\n", expr->symtree->n.sym->name); */
    rcall->routine = pdbGetRoutine(expr->symtree->n.sym);
    rcall->routineID = rcall->routine->id;
  } else {
/*     printf ("Hmm, what is this?\n"); */
  }
  rcall->location.file = tau_so_num;
  rcall->location.line = expr->where.pdbStartLine;
  rcall->location.column = expr->where.nextc - expr->where.lb->line +1;
  rcall->nextLink = NULL;

  if (pdbFirstRcall == NULL) {
    pdbFirstRcall = rcall;
    pdbLastRcall = rcall;
  } else {
    pdbLastRcall->nextLink = rcall;
  }
  pdbLastRcall = rcall;

  return rcall;
}

pdbRcall *createRcallS(gfc_code *code) {

  pdbRcall *rcall;
  rcall = (pdbRcall *) gfc_getmem (sizeof(pdbRcall));
  rcall->routine = pdbGetRoutine(code->resolved_sym);
  rcall->routineID = rcall->routine->id;
  rcall->location.file = tau_so_num;
  rcall->location.line = code->loc.pdbStartLine;
  rcall->location.column = code->loc.pdbStartColumn;
  rcall->nextLink = NULL;

  if (pdbFirstRcall == NULL) {
    pdbFirstRcall = rcall;
    pdbLastRcall = rcall;
  } else {
    pdbLastRcall->nextLink = rcall;
  }
  pdbLastRcall = rcall;

  return rcall;
}



#define PDB_RETURN 0
#define PDB_STOP 1
#define PDB_ASSIGN 2
#define PDB_CONTINUE 3
#define PDB_IO 4
#define PDB_DO 5
#define PDB_IF 6
#define PDB_ALLOCATE 7
#define PDB_DEALLOCATE 8
#define PDB_GOTO 9
#define PDB_LABEL 10
#define PDB_EXIT 11
#define PDB_CYCLE 12
#define PDB_ARITHIF 13
#define PDB_ENTRY 14
#define PDB_PAUSE 15
#define PDB_LABELASSIGN 16
#define PDB_POINTERASSIGN 17
#define PDB_SELECT 18
#define PDB_CASE 19
#define PDB_WHERE 20
#define PDB_FORALL 21
#define PDB_CALL 22
#define PDB_SINGLEIF 23
const char *stkinds[24] = {"freturn", "fstop", "fassign", "fcontinue", "fio", 
			 "fdo", "fif", "fallocate", "fdeallocate", "fgoto", "flabel", 
			 "fexit", "fcycle", "farithif", "fentry", "fpause", "flabelassign", 
			 "fpointerassign", "fselect", "fcase", "fwhere", "fforall", "fcall", "fsingle_if"};



pdbStatement *firstStatement;
pdbStatement *lastStatement;
pdbStatement *pdbPrev;
pdbStatement *pdbUpper;
pdbStatement *pdbExtra;

int numStatements;

extern int tau_ro_num;
extern int tau_ty_num;

static void pdb_visit_code_node (int level, gfc_code * c);
static void gfc_show_expr (gfc_expr *);

#define code_indent(A,B) 

/* Show an expression.  */


static char *myStrDup(const char *src) {
  char *dest;
  dest = gfc_getmem(strlen(src)+1);
  memcpy(dest,src,strlen(src)+1);
  return dest;
}

static const char *genName(gfc_symbol *sym) {
  char *name, *tmp;
  gfc_namespace *ns;
  int gen;
  unsigned int i;

  if (!sym) {
    return "unknown";
  }

  gen = 0;
  name = (char *) sym->name;

  if (sym->module) {
      tmp = name;
      name = (char*) gfc_getmem(strlen(sym->module) + strlen(tmp) + 3);
      sprintf (name, "%s::%s", sym->module, tmp);
      gen = 1;

  }

  for (ns = sym->ns; ns; ns = ns->parent) {
    gfc_symbol *proc_name = ns->proc_name;

    if (proc_name != sym && proc_name) {
      tmp = name;
      name = (char*) gfc_getmem(strlen(proc_name->name) + strlen(tmp) + 3);
      sprintf (name, "%s::%s", proc_name->name, tmp);
      if (gen) {
	gfc_free (tmp);
      }
      gen = 1;

    }
  }

  for (i=0;i<strlen(name);i++) {
    name[i] = toupper(name[i]);
  }

  return name;
}

static pdbRoutine *pdbGetRoutine(gfc_symbol *sym) {
  pdbRoutine *routine;
  const char *searchname;
  const char *name;


/*   searchname = sym->name; */
  searchname = genName(sym);

  for (routine = pdbFirstRoutine; routine; routine = routine->nextLink) {
    name = routine->name;

    if (strcmp(name,searchname) == 0) {
      return routine;
    }
  }

  routine = (pdbRoutine *) gfc_getmem (sizeof(pdbRoutine));
  routine->id = tau_ro_num++;
  routine->name = searchname;
  routine->sym = sym;
  routine->type = NULL;
/*   routine->appearsInThisFile = 1; */
  /* probably needs to be adjusted for modules, etc */

  if (pdbFirstRoutine == NULL) {
    pdbFirstRoutine = routine;
    pdbLastRoutine = routine;
  } else {
    pdbLastRoutine->nextLink = routine;
  }
  pdbLastRoutine = routine;

  return routine;
}



static const char *kindToString(int kind) {
  return stkinds[kind];
}

static void outputRcall(pdbRcall *rcall) {
  fprintf (pdbfile, "rcall ro#%d no so#%d %d %d\n", rcall->routineID, rcall->location.file, 
	   rcall->location.line, rcall->location.column);
}

static void outputStatement(pdbStatement *stmt) {

  /* TEMP: don't output anything but return and stop, so I can use it with old PDT */

#ifdef DISABLE_STATEMENTS
  if (stmt->kind != PDB_RETURN && stmt->kind != PDB_STOP) {
    return;
  }
#endif

  fprintf (pdbfile, "rstmt st#%d %s ", stmt->id, kindToString(stmt->kind));

  if (stmt->start.line < 0) {
    fprintf (pdbfile,"NULL 0 0 ");
  } else {
    fprintf (pdbfile,"so#%d %d %d ", stmt->start.file, stmt->start.line, stmt->start.column); 
  }
  
  if (stmt->end.line < 0) {
    fprintf (pdbfile,"NULL 0 0 ");
  } else {
    fprintf (pdbfile,"so#%d %d %d ", stmt->end.file, stmt->end.line, stmt->end.column); 
  }

  if (stmt->nextstmt) {
    fprintf (pdbfile,"st#%d ", stmt->nextstmt->id);
  } else {
    fprintf (pdbfile,"NA ");
  }

  if (stmt->downstmt) {
    fprintf (pdbfile,"st#%d ", stmt->downstmt->id);
  } else {
    fprintf (pdbfile,"NA ");
  }
  
  if (stmt->extrastmt) {
    fprintf (pdbfile,"st#%d ", stmt->extrastmt->id);
  }

  fprintf (pdbfile, "\n");
  
/*   fprintf (pdbfile, "rstmt st#%d %s so#%d %d %d so#%d 0 0 st#%d NA\n", tau_st_num, tau_so_num, c->loc.lb->linenum, theColumn, tau_st_num+1); *\/ */
 
}

static void newFunction(void) {
  numStatements = 0;

  pdbPrev = NULL;
  pdbUpper = NULL;
  pdbExtra = NULL;
  firstStatement = NULL;
  lastStatement = NULL;

  pdbFirstRcall = NULL;
  pdbLastRcall = NULL;

}

static void
gfc_show_actual_arglist (gfc_actual_arglist * a)
{

  gfc_status ("(");

  for (; a; a = a->next)
    {
      gfc_status_char ('(');
      if (a->name != NULL)
	gfc_status ("%s = ", a->name);
      if (a->expr != NULL)
	gfc_show_expr (a->expr);
      else
	gfc_status ("(arg not-present)");

      gfc_status_char (')');
      if (a->next != NULL)
	gfc_status (" ");
    }

  gfc_status (")");
}

static void
gfc_show_array_ref (gfc_array_ref * ar)
{
  int i;

  gfc_status_char ('(');

  switch (ar->type)
    {
    case AR_FULL:
      gfc_status ("FULL");
      break;

    case AR_SECTION:
      for (i = 0; i < ar->dimen; i++)
	{
	  /* There are two types of array sections: either the
	     elements are identified by an integer array ('vector'),
	     or by an index range. In the former case we only have to
	     print the start expression which contains the vector, in
	     the latter case we have to print any of lower and upper
	     bound and the stride, if they're present.  */
  
	  if (ar->start[i] != NULL)
	    gfc_show_expr (ar->start[i]);

	  if (ar->dimen_type[i] == DIMEN_RANGE)
	    {
	      gfc_status_char (':');

	      if (ar->end[i] != NULL)
		gfc_show_expr (ar->end[i]);

	      if (ar->stride[i] != NULL)
		{
		  gfc_status_char (':');
		  gfc_show_expr (ar->stride[i]);
		}
	    }

	  if (i != ar->dimen - 1)
	    gfc_status (" , ");
	}
      break;

    case AR_ELEMENT:
      for (i = 0; i < ar->dimen; i++)
	{
	  gfc_show_expr (ar->start[i]);
	  if (i != ar->dimen - 1)
	    gfc_status (" , ");
	}
      break;

    case AR_UNKNOWN:
      gfc_status ("UNKNOWN");
      break;

    default:
      gfc_internal_error ("gfc_show_array_ref(): Unknown array reference");
    }

  gfc_status_char (')');
}

static void
gfc_show_ref (gfc_ref * p)
{

  for (; p; p = p->next)
    switch (p->type)
      {
      case REF_ARRAY:
	gfc_show_array_ref (&p->u.ar);
	break;

      case REF_COMPONENT:
	gfc_status (" %% %s", p->u.c.component->name);
	break;

      case REF_SUBSTRING:
	gfc_status_char ('(');
	gfc_show_expr (p->u.ss.start);
	gfc_status_char (':');
	gfc_show_expr (p->u.ss.end);
	gfc_status_char (')');
	break;

      default:
	gfc_internal_error ("gfc_show_ref(): Bad component code");
      }
}

static void
gfc_show_constructor (gfc_constructor * c)
{

  for (; c; c = c->next)
    {
      if (c->iterator == NULL)
	gfc_show_expr (c->expr);
      else
	{
	  gfc_status_char ('(');
	  gfc_show_expr (c->expr);

	  gfc_status_char (' ');
	  gfc_show_expr (c->iterator->var);
	  gfc_status_char ('=');
	  gfc_show_expr (c->iterator->start);
	  gfc_status_char (',');
	  gfc_show_expr (c->iterator->end);
	  gfc_status_char (',');
	  gfc_show_expr (c->iterator->step);

	  gfc_status_char (')');
	}

      if (c->next != NULL)
	gfc_status (" , ");
    }
}

static void
gfc_show_expr (gfc_expr * p)
{
  const char *c;
  int i;

  if (p == NULL)
    {
      gfc_status ("()");
      return;
    }

  switch (p->expr_type)
    {
    case EXPR_SUBSTRING:
      c = p->value.character.string;

      for (i = 0; i < p->value.character.length; i++, c++)
	{
	  if (*c == '\'')
	    gfc_status ("''");
	  else
	    gfc_status ("%c", *c);
	}

      gfc_show_ref (p->ref);
      break;

    case EXPR_STRUCTURE:
      gfc_status ("%s(", p->ts.derived->name);
      gfc_show_constructor (p->value.constructor);
      gfc_status_char (')');
      break;

    case EXPR_ARRAY:
      gfc_status ("(/ ");
      gfc_show_constructor (p->value.constructor);
      gfc_status (" /)");

      gfc_show_ref (p->ref);
      break;

    case EXPR_NULL:
      gfc_status ("NULL()");
      break;

    case EXPR_CONSTANT:
      switch (p->ts.type)
	{
	case BT_INTEGER:
/* 	  mpz_out_str (stdout, 10, p->value.integer); */
/* 	  gfc_status("%d", p->value.integer); */

	  if (p->ts.kind != gfc_default_integer_kind)
	    gfc_status ("_%d", p->ts.kind);
	  break;

	case BT_LOGICAL:
	  if (p->value.logical)
	    gfc_status (".true.");
	  else
	    gfc_status (".false.");
	  break;

	case BT_REAL:
/* 	  mpfr_out_str (stdout, 10, 0, p->value.real, GFC_RND_MODE); */
/* 	  if (p->ts.kind != gfc_default_real_kind) */
/* 	    gfc_status ("_%d", p->ts.kind); */
	  break;

	case BT_CHARACTER:
	  c = p->value.character.string;

	  gfc_status_char ('\'');

	  for (i = 0; i < p->value.character.length; i++, c++)
	    {
	      if (*c == '\'')
		gfc_status ("''");
	      else
		gfc_status_char (*c);
	    }

	  gfc_status_char ('\'');

	  break;

	case BT_COMPLEX:
	  gfc_status ("(complex ");

	  mpfr_out_str (stdout, 10, 0, p->value.complex.r, GFC_RND_MODE);
	  if (p->ts.kind != gfc_default_complex_kind)
	    gfc_status ("_%d", p->ts.kind);

	  gfc_status (" ");

	  mpfr_out_str (stdout, 10, 0, p->value.complex.i, GFC_RND_MODE);
	  if (p->ts.kind != gfc_default_complex_kind)
	    gfc_status ("_%d", p->ts.kind);

	  gfc_status (")");
	  break;

	default:
	  gfc_status ("???");
	  break;
	}

      break;

    case EXPR_VARIABLE:
      if (p->symtree->n.sym->ns && p->symtree->n.sym->ns->proc_name)
	gfc_status ("%s:", p->symtree->n.sym->ns->proc_name->name);
      gfc_status ("%s", p->symtree->n.sym->name);
      gfc_show_ref (p->ref);
      break;

    case EXPR_OP:
      gfc_status ("(");
      switch (p->value.op.operator)
	{
	case INTRINSIC_UPLUS:
	  gfc_status ("U+ ");
	  break;
	case INTRINSIC_UMINUS:
	  gfc_status ("U- ");
	  break;
	case INTRINSIC_PLUS:
	  gfc_status ("+ ");
	  break;
	case INTRINSIC_MINUS:
	  gfc_status ("- ");
	  break;
	case INTRINSIC_TIMES:
	  gfc_status ("* ");
	  break;
	case INTRINSIC_DIVIDE:
	  gfc_status ("/ ");
	  break;
	case INTRINSIC_POWER:
	  gfc_status ("** ");
	  break;
	case INTRINSIC_CONCAT:
	  gfc_status ("// ");
	  break;
	case INTRINSIC_AND:
	  gfc_status ("AND ");
	  break;
	case INTRINSIC_OR:
	  gfc_status ("OR ");
	  break;
	case INTRINSIC_EQV:
	  gfc_status ("EQV ");
	  break;
	case INTRINSIC_NEQV:
	  gfc_status ("NEQV ");
	  break;
	case INTRINSIC_EQ:
	  gfc_status ("= ");
	  break;
	case INTRINSIC_NE:
	  gfc_status ("<> ");
	  break;
	case INTRINSIC_GT:
	  gfc_status ("> ");
	  break;
	case INTRINSIC_GE:
	  gfc_status (">= ");
	  break;
	case INTRINSIC_LT:
	  gfc_status ("< ");
	  break;
	case INTRINSIC_LE:
	  gfc_status ("<= ");
	  break;
	case INTRINSIC_NOT:
	  gfc_status ("NOT ");
	  break;

	default:
	  gfc_internal_error
	    ("gfc_show_expr(): Bad intrinsic in expression!");
	}

      gfc_show_expr (p->value.op.op1);

      if (p->value.op.op2)
	{
	  gfc_status (" ");
	  gfc_show_expr (p->value.op.op2);
	}

      gfc_status (")");
      break;

    case EXPR_FUNCTION:
      if (p->value.function.name == NULL) {
	gfc_status ("%s[", p->symtree->n.sym->name);
/* 	printf ("[RCALL] %s (%p)\n", p->symtree->n.sym->name, p->symtree->n.sym); */
	gfc_show_actual_arglist (p->value.function.actual);
      } else {
/* 	printf ("    TRY [RCALL-F] %s (%p) at (%d:%d)\n", p->value.function.name, p,  */
/* 		p->where.pdbStartLine, p->where.nextc - p->where.lb->line +1); */

#ifdef PDB_ENABLE_RCALL
	pdbRcall *rcall = createRcall(p);
#endif

      }

      break;

    default:
      gfc_internal_error ("gfc_show_expr(): Don't know how to show expr");
    }
}


static void pdb_visit_code (int level, gfc_code * c) {
  for (; c; c = c->next) {
    pdb_visit_code_node (level, c);
  }
}



static pdbStatement *newStatement(int kind) {
  pdbStatement *stmt = (pdbStatement *) gfc_getmem (sizeof(pdbStatement));

  stmt->kind = kind;

#ifdef DISABLE_STATEMENTS
  if (stmt->kind != PDB_RETURN && stmt->kind != PDB_STOP) {
    stmt->id = -1;
  } else {
#endif
    stmt->id = numStatements++;
#ifdef DISABLE_STATEMENTS
  }
#endif

  stmt->start.file = -1;
  stmt->start.line = -1;
  stmt->start.column = -1;
  stmt->end.file = -1;
  stmt->end.line = -1;
  stmt->end.column = -1;

  stmt->nextstmt = NULL;
  stmt->downstmt = NULL;
  stmt->extrastmt = NULL;
  stmt->nextLink = NULL;
  stmt->label = NULL;
  stmt->code = NULL;


#ifdef DISABLE_STATEMENTS
  if (stmt->kind != PDB_RETURN && stmt->kind != PDB_STOP) {
    return stmt;
  }
#endif

  if (!firstStatement) {
    firstStatement = stmt;
    lastStatement = stmt;
  } else {
    lastStatement->nextLink = stmt;
    lastStatement = stmt;
  }

  return stmt;
}


static pdbStatement *createStatement(int kind, locus loc) {
  int startColumn, endColumn;
  pdbStatement *stmt;

  startColumn = loc.pdbStartColumn;
  if (loc.lb) {
    startColumn -= loc.lb->tab;
  }

  endColumn = loc.pdbEndColumn;
  if (loc.lb) {
    endColumn -= loc.lb->tab;
  }

  stmt = newStatement(kind);
  stmt->start.file = tau_so_num;
  stmt->start.line = loc.pdbStartLine;
  stmt->start.column = startColumn;
  stmt->end.file = tau_so_num;
  stmt->end.line = loc.pdbEndLine;
  stmt->end.column = endColumn;
  return stmt;
}



static void insertStatement(pdbStatement *stmt) {

#ifdef DISABLE_STATEMENTS
  if (stmt->kind != PDB_RETURN && stmt->kind != PDB_STOP) {
    return;
  }
#endif

  if (pdbPrev) {
    pdbPrev->nextstmt = stmt;
  }

  if (pdbUpper) {
    pdbUpper->downstmt = stmt;
    pdbUpper = NULL;
  }

  if (pdbExtra) {
    pdbExtra->extrastmt = stmt;
    pdbExtra = NULL;
  }
  
  pdbPrev = stmt;
}


static void pdb_visit_code_node (int level, gfc_code * c) {
  gfc_code *d;

  int theColumn, endColumn, endLine;
  pdbStatement *stmt, *stmt2, *if_stmt, *where_stmt;


  code_indent (level, c->here);

  if (c->here) {
      stmt = createStatement(PDB_LABEL,c->here->where);
      stmt->end.line = -1;
      stmt->label = c->here;
      insertStatement(stmt);
  }

  endLine = c->loc.pdbEndLine;


  theColumn = c->loc.pdbStartColumn;
  
  if (c->loc.lb) {
    theColumn -= c->loc.lb->tab;
  }

  

  endColumn = c->loc.pdbEndColumn;
  if (c->loc.lb) {
    endColumn -= c->loc.lb->tab;
  }


  /* process expressions, to find rcall's */
  if (c->expr) gfc_show_expr (c->expr);
  if (c->expr2) gfc_show_expr (c->expr2);


/*   if (c->op != EXEC_STOP || c->op != EXEC_RETURN) { */
/*     return; */
/*   } */

  switch (c->op)
    {
    case EXEC_NOP:
      gfc_status ("NOP");
      break;

    case EXEC_CONTINUE:
      stmt = createStatement(PDB_CONTINUE,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_ENTRY:
      /* Multiple program unit entries not currently supported */
      stmt = createStatement(PDB_ENTRY,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_ASSIGN:
      stmt = createStatement(PDB_ASSIGN,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_LABEL_ASSIGN:
      stmt = createStatement(PDB_LABELASSIGN,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_POINTER_ASSIGN:
      stmt = createStatement(PDB_POINTERASSIGN,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_GOTO:
      stmt = createStatement(PDB_GOTO,c->loc);
      stmt->label = c->label;
      insertStatement(stmt);
      break;

    case EXEC_CALL:
      stmt = createStatement(PDB_CALL,c->loc);
      stmt->end.line = c->loc.pdbAltEndLine;
      stmt->end.column = c->loc.pdbAltEndColumn;
      insertStatement(stmt);

#ifdef PDB_ENABLE_RCALL
      {
	pdbRcall *rcall = createRcallS(c);
      }
#endif
      
      /* hmm */
      gfc_show_actual_arglist (c->ext.actual);
      break;

    case EXEC_RETURN:
      stmt = createStatement(PDB_RETURN,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_PAUSE:
      stmt = createStatement(PDB_PAUSE,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_STOP:
      stmt = createStatement(PDB_STOP,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_ARITHMETIC_IF:
      stmt = createStatement(PDB_ARITHIF,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_IF:
      if (c->single_if) {
	stmt = createStatement(PDB_SINGLEIF,c->loc);
      } else {
	stmt = createStatement(PDB_IF,c->loc);
      }
      insertStatement(stmt);

      d = c->block;

      /* now visit the 'down' block */
      pdbPrev = NULL; pdbUpper = stmt; pdbExtra = NULL;
      pdb_visit_code (level + 1, d->next);

      if_stmt = stmt;

      /* there may be an else or else if */
      d = d->block;
      for (; d; d = d->block) {
	
	/* set the extra */
	pdbPrev = NULL; pdbUpper = NULL; pdbExtra = if_stmt;
	
	if (d->expr == NULL) {
/* 	  gfc_status ("ELSE\n"); */
	} else {
/* 	  gfc_status ("ELSE IF "); */
	  
	  stmt2 = createStatement(PDB_IF,d->loc);
	  insertStatement(stmt2);
	  
	  pdbPrev = NULL; pdbUpper = stmt2; pdbExtra = NULL;
	  if_stmt = stmt2;
	}
	
	pdb_visit_code (level + 1, d->next);
      }
      
      /* Set the prev to the top IF */
      pdbUpper = NULL; pdbExtra = NULL; pdbPrev = stmt;

      break;

    case EXEC_SELECT:

      stmt = createStatement(PDB_SELECT,c->loc);
      stmt->end.line = c->loc.pdbAltEndLine;
      stmt->end.column = c->loc.pdbAltEndColumn;
      insertStatement(stmt);

      /* set the SELECT to be the upper of the first case */
      pdbPrev = NULL; pdbUpper = stmt; pdbExtra = NULL;

      d = c->block;

      for (; d; d = d->block) {
	stmt2 = createStatement(PDB_CASE,d->ext.case_list->where);
	insertStatement(stmt2);
	
	/* set the CASE to be the upper of the body */
	pdbPrev = NULL; pdbUpper = stmt2; pdbExtra = NULL;
	
	pdb_visit_code (level + 1, d->next);
	
	/* set the CASE to be the prevr of the next case */
	pdbPrev = stmt2; pdbUpper = NULL; pdbExtra = NULL;
	  
      }

      /* set the SELECT to be the prev of the next statement */
      pdbPrev = stmt; pdbUpper = NULL; pdbExtra = NULL;

      break;

    case EXEC_WHERE:
      stmt = createStatement(PDB_WHERE,c->loc);
      insertStatement(stmt);

      d = c->block;


      /* now visit the 'down' block */
      pdbPrev = NULL; pdbUpper = stmt; pdbExtra = NULL;
      pdb_visit_code (level + 1, d->next);

      where_stmt = stmt;

      for (d = d->block; d; d = d->block) {

	/* set the extra */
	pdbPrev = NULL; pdbUpper = NULL; pdbExtra = where_stmt;

	stmt2 = createStatement(PDB_WHERE,d->loc);
	insertStatement(stmt2);

	pdbPrev = NULL; pdbUpper = stmt2; pdbExtra = NULL;
	where_stmt = stmt2;

	pdb_visit_code (level + 1, d->next);
      }

      /* Set the prev to the top WHERE */
      pdbUpper = NULL; pdbExtra = NULL; pdbPrev = stmt;

      break;


    case EXEC_FORALL:
      stmt = createStatement(PDB_FORALL,c->loc);
      stmt->end.line = c->loc.pdbAltEndLine;
      stmt->end.column = c->loc.pdbAltEndColumn;
      insertStatement(stmt);

      pdbPrev = NULL; pdbUpper = stmt; pdbExtra = NULL;
      pdb_visit_code (level + 1, c->block->next);
      pdbPrev = stmt; pdbUpper = NULL; pdbExtra = NULL;

      break;

    case EXEC_DO:
    case EXEC_DO_WHILE:
      // there are EXEC_DO's generated for loop io statements, such as:
      //   write(*,*) (title(i), i=1,3,1)
      // I set their pdbLoopEndLine to -1
      if (c->loc.pdbLoopEndLine >= 0) {
	stmt = newStatement(PDB_DO);
	stmt->start.file = tau_so_num;
	stmt->start.line = c->loc.pdbStartLine;
	stmt->start.column = theColumn;
	stmt->end.file = tau_so_num;
	stmt->end.line = c->loc.pdbLoopEndLine;
	stmt->end.column = c->loc.pdbLoopEndColumn;
	
	stmt->code = c;
	insertStatement(stmt);
	
	pdbPrev = NULL; pdbUpper = stmt; pdbExtra = NULL;
	pdb_visit_code (level + 1, c->block->next);
	pdbPrev = stmt; pdbUpper = NULL; pdbExtra = NULL;
      }
      break;

    case EXEC_CYCLE:
      stmt = createStatement(PDB_CYCLE,c->loc);
      stmt->code = c;
      insertStatement(stmt);
      break;

    case EXEC_EXIT:
      stmt = createStatement(PDB_EXIT,c->loc);
      stmt->code = c;
      insertStatement(stmt);
      break;

    case EXEC_ALLOCATE:
      stmt = createStatement(PDB_ALLOCATE,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_DEALLOCATE:
      stmt = createStatement(PDB_DEALLOCATE,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_OPEN:
    case EXEC_CLOSE:
    case EXEC_BACKSPACE:
    case EXEC_ENDFILE:
    case EXEC_REWIND:
    case EXEC_INQUIRE:
    case EXEC_IOLENGTH:
    case EXEC_READ:
    case EXEC_WRITE:
      stmt = createStatement(PDB_IO,c->loc);
      insertStatement(stmt);
      break;

    case EXEC_TRANSFER:
/*       gfc_status ("TRANSFER "); */
/*       gfc_show_expr (c->expr); */
      break;

    case EXEC_DT_END:
      break;

    default:
      gfc_internal_error ("pdb_visit_code_node(): Bad statement code");
    }
}



static void
gfc_show_symbol (gfc_symbol * sym)
{
/*   gfc_formal_arglist *formal; */
/*   gfc_interface *intr; */

  if (sym == NULL)
    return;
 

/*   if (sym->attr.flavor == FL_PROCEDURE) { */

/*   if (sym->attr.flavor == FL_MODULE) { */
/*     printf ("Found Module: %s\n", sym->name); */
/*   } */

  if (sym->attr.flavor == FL_PROCEDURE || sym->attr.flavor == FL_PROGRAM) {


    pdbGetRoutine(sym);
/*     printf ("Found PROCEDURE: %s\n", sym->name); */
    
/*     sym->pdbRoutineID = tau_ro_num++; */

/*     if (sym->formal) { */
/*       printf ("Formal arglist:\n"); */
      
/*       for (formal = sym->formal; formal; formal = formal->next) { */
/* 	if (formal->sym != NULL) */
/* 	  printf (" %s\n", formal->sym->name); */
/* 	else */
/* 	  printf (" [Alt Return]\n"); */
/*       } */
/*     } */
    
  }

/*   gfc_status ("symbol %s ", sym->name); */
/*   gfc_show_typespec (&sym->ts); */
/*   gfc_show_attr (&sym->attr); */

/*   if (sym->value) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("value: "); */
/*       gfc_show_expr (sym->value); */
/*     } */

/*   if (sym->as) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("Array spec:"); */
/*       gfc_show_array_spec (sym->as); */
/*     } */

/*   if (sym->generic) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("Generic interfaces:"); */
/*       for (intr = sym->generic; intr; intr = intr->next) */
/* 	gfc_status (" %s", intr->sym->name); */
/*     } */

/*   if (sym->result) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("result: %s", sym->result->name); */
/*     } */

/*   if (sym->components) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("components: "); */
/*       gfc_show_components (sym); */
/*     } */

/*   if (sym->formal) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("Formal arglist:"); */

/*       for (formal = sym->formal; formal; formal = formal->next) */
/*         { */
/*           if (formal->sym != NULL) */
/*             gfc_status (" %s", formal->sym->name); */
/*           else */
/*             gfc_status (" [Alt Return]"); */
/*         } */
/*     } */

/*   if (sym->formal_ns) */
/*     { */
/*       show_indent (); */
/*       gfc_status ("Formal namespace"); */
/*       gfc_show_namespace (sym->formal_ns); */
/*     } */

/*   gfc_status_char ('\n'); */
}


static void pdb_symtree (gfc_symtree * st) {

/*   printf ("visit symbol: %s  Ambig %d\n", st->name, st->ambiguous); */

  if (st->n.sym->ns != gfc_current_ns) {
/*     printf (" from namespace %s\n", st->n.sym->ns->proc_name->name); */
  } else {
    gfc_show_symbol (st->n.sym);
  }
}


pdbType *createType(gfc_symbol *sym);


void insertType(pdbType *type);
void insertType(pdbType *type) {
/*   int i; */

/*   type->name = myStrDup(type->name); */
/*   printf ("%d : %s (%d)\n", type->id, type->name, strlen(type->name)); */
/*   for (i=0;i<strlen(type->name);i++) { */
/*     type->name[i] = toupper(type->name[i]);  */
/*   } */


  /* insert */
  if (pdbFirstType == NULL) {
    pdbFirstType = type;
    pdbLastType = type;
  } else {
    pdbLastType->nextLink = type;
  }
  pdbLastType = type;
}

void insertDerived(gfc_symbol *sym, pdbType *type);
void insertDerived(gfc_symbol *sym, pdbType *type) {
  derivedListNode *node = gfc_getmem (sizeof(derivedListNode));
  node->sym = sym;
  node->type = type;
  node->next = NULL;
  
  if (derivedFirst == NULL) {
    derivedFirst = node;
    derivedLast = node;
  } else {
    node->next = derivedFirst;
    derivedFirst = node;
  }
}


pdbType *getDerived(gfc_symbol *sym) {
  derivedListNode *node = derivedFirst;
  while (node) {
    if (node->sym == sym) {
      return node->type;
    }
    node = node->next;
  }
  return NULL;
}



pdbType *createTType(gfc_typespec ts, gfc_array_spec *as, int pointer);
pdbType *createArrayType(gfc_typespec ts, gfc_array_spec *as);

pdbType *createArrayType(gfc_typespec ts, gfc_array_spec *as) {
  pdbType *type;

  type = (pdbType *) gfc_getmem (sizeof(pdbType));
  type->id = tau_ty_num++;
  sprintf(type->idString, "ty#%d", type->id);
  type->name = (char *) "some array type";
  type->baseType = createTType(ts, NULL, 0);
 
  char *name = (char *) gfc_getmem(strlen(type->baseType->name) + 20);
  sprintf(name, "%s[]", type->baseType->name);
  type->name = name;
  type->kind = KIND_ARRAY;
  type->as = *as;
  insertType(type);
  return type;
}

pdbType *createPointerType(gfc_typespec ts, gfc_array_spec *as);
pdbType *createPointerType(gfc_typespec ts, gfc_array_spec *as) {
  pdbType *type;

  type = (pdbType *) gfc_getmem (sizeof(pdbType));
  type->id = tau_ty_num++;
  sprintf(type->idString, "ty#%d", type->id);
  type->name = (char *)"some pointer type";
  type->baseType = createTType(ts, as, 0);

  char *name = (char *) gfc_getmem(strlen(type->baseType->name) + 20);
  sprintf(name, "*%s", type->baseType->name);
  type->name = name;

  type->kind = KIND_POINTER;
  insertType(type);
  return type;
}

pdbType *createTType(gfc_typespec ts, gfc_array_spec *as, int pointer) {
  pdbType *type;

  if (pointer) {
    return createPointerType(ts, as);
  }

  if (as) {
    return createArrayType(ts, as);
  }


  type = (pdbType *) gfc_getmem (sizeof(pdbType));

  if (ts.type == BT_DERIVED) {
    pdbType *findType;
    findType = getDerived(ts.derived);
    if (findType != NULL) {
      return findType;
    } else {
      insertDerived(ts.derived, type);
    }
  }


  type->id = tau_ty_num++;
  type->name = "sometype";
  
  sprintf(type->idString, "ty#%d", type->id);
  switch (ts.type) {
  case BT_UNKNOWN:
    type->kind = KIND_VOID;
    type->name = "void";
    break;
  case BT_INTEGER:
    type->kind = KIND_INT;
    switch (ts.kind) {
    case 2:
      type->ikind = IKIND_SHORT;
      type->name = "int16";
      break;
    case 8:
      type->ikind = IKIND_LONGLONG;
      type->name = "int64";
      break;
    case 4:
    default: /* default to 4 */
      type->ikind = IKIND_INT;
      type->name = "int32";
      break;
    }
    break;
  case BT_REAL:
    type->kind = KIND_FLOAT;

    switch (ts.kind) {
    case 4:
      type->fkind = FKIND_FLOAT;
      type->name = "float";
      break;
    case 8:
      type->fkind = FKIND_DOUBLE;
      type->name = "double";
      break;
    case 16:
      type->fkind = FKIND_LONGDOUBLE;
      type->name = "quad";
    }
    break;
  case BT_COMPLEX:
    type->kind = KIND_COMPLEX;
    type->name = "complex";
    switch (ts.kind) {
    case 4:
      type->fkind = FKIND_FLOAT;
      type->name = "complex32";
      break;
    case 8:
      type->fkind = FKIND_DOUBLE;
      type->name = "complex64";
      break;
    case 16:
      type->fkind = FKIND_LONGDOUBLE;
      type->name = "complex128";
    }
    break;
  case BT_LOGICAL:
    type->kind = KIND_LOGICAL;
    type->name = "logical";
    switch (ts.kind) {
    case 1:
      type->ikind = IKIND_CHAR;
      type->name = "logical8";
      break;
    case 2:
      type->ikind = IKIND_SHORT;
      type->name = "logical16";
      break;
    case 8:
      type->ikind = IKIND_LONGLONG;
      type->name = "logical64";
      break;
    case 4:
      type->ikind = IKIND_INT;
      type->name = "logical32";
      break;
    }
    break;
  case BT_CHARACTER:
    type->kind = KIND_CHAR;
    type->name = "character";
    break;
  case BT_DERIVED:
    sprintf(type->idString, "gr#%d", type->id);

    {
      gfc_symbol *sym = ts.derived;
      pdbArg *arg;
      pdbArg *lastArg = NULL;
      gfc_component *comp;
      type->args = NULL;
      type->name = sym->name;
      type->kind = KIND_DERIVED;
      
      for (comp = sym->components ; comp; comp = comp->next) {
	
	arg = (pdbArg *) gfc_getmem(sizeof(pdbArg));
	arg->name = comp->name;
	arg->type = createTType(comp->ts,comp->as,comp->pointer);
	arg->type->access = sym->component_access;
	/* insert */
	if (type->args == NULL) {
	  type->args = arg;
	  lastArg = arg;
	} else {
	  lastArg->nextLink = arg;
	}
	lastArg = arg;
	
      }
    }

    break;
  case BT_PROCEDURE:
    break;
  case BT_HOLLERITH:
    break;
    
  default:
    break;
  }


  insertType(type);
  return type;
}

pdbType *createType(gfc_symbol *sym) {
  pdbType *type;
  gfc_formal_arglist *formal;
  pdbArg *lastArg, *arg;


  if (sym->attr.flavor == FL_PROCEDURE || sym->attr.flavor == FL_PROGRAM) {
    int namelength;
    int first;

    namelength = 0;
    type = (pdbType *) gfc_getmem (sizeof(pdbType));
    
    type->kind = KIND_NA;
    type->id = tau_ty_num++;
    sprintf(type->idString, "ty#%d", type->id);
    type->kind = KIND_FUNC;
    type->returnType = createTType(sym->ts,sym->as,sym->attr.pointer);


    type->args = NULL;
    lastArg = NULL;
    for (formal = sym->formal; formal; formal = formal->next) {
      if (formal->sym != NULL) {
	
	arg = (pdbArg *) gfc_getmem(sizeof(pdbArg));
	arg->name = formal->sym->name;
	arg->type = createType(formal->sym);
	namelength = namelength + strlen(arg->type->name) + 2;
	/* insert */
	if (type->args == NULL) {
	  type->args = arg;
	  lastArg = arg;
	} else {
	  lastArg->nextLink = arg;
	}
	lastArg = arg;

      } else {
	printf ("Error in createType!\n");
      }
    }

    namelength = namelength + strlen(type->returnType->name);

    /* count string length */
    first = 1;
    for (arg = type->args; arg ; arg = arg->nextLink) {
      namelength = namelength + strlen(arg->name) + 5;
    }

    char *name = (char *) gfc_getmem (namelength + 20); /* +20 for commas, parens, etc */
    sprintf (name,"%s (", type->returnType->name);

    first = 1;
    for (arg = type->args; arg ; arg = arg->nextLink) {
      if (!first) {
	strcat (name, ", ");
      }
      first = 0;
      strcat (name, arg->name);
    }
    strcat (name, ")");
    type->name = name;

      
    insertType(type);


  } else {

    type = createTType(sym->ts,sym->as,sym->attr.pointer);
  }


  return type;
  
}

void pdt_output_pdb (gfc_namespace * ns);
void pdt_output_pdb (gfc_namespace * ns) {
  int theColumn;
  locus rstart, end;
  pdbStatement *stmt, *tmpstmt;
/*   pdbRcall *rcall; */
  pdbRoutine *routine;

/*   printf("\n                  -= pdt_output_pdb =-\n\n"); */

  if (ns != NULL) {
    

      /* traverse the tree to get the functions */
      gfc_traverse_symtree (ns->sym_root, pdb_symtree);
    
    if (ns->tau_progunit != NULL) {
/*       printf ("Generating PDB for Program Unit \"%s\"\n", ns->tau_progunit->name); */

      newFunction();

      routine = pdbGetRoutine(ns->tau_progunit);
/*       printf ("ROUTINE NAME = %s\n", routine->name); */

      pdb_visit_code(0, ns->code);

      
      /* Now, let's resolve goto/labels, using n^2 algo!!! */
      for (stmt = firstStatement; stmt; stmt = stmt->nextLink) {
	if (stmt->kind == PDB_GOTO) {
	  for (tmpstmt = firstStatement; tmpstmt; tmpstmt = tmpstmt->nextLink) {
	    if (tmpstmt->kind == PDB_LABEL && stmt->label == tmpstmt->label) {
	      stmt->extrastmt = tmpstmt;
	    }
	  }
	}
      }


      /* Now, let's resolve exit/cycle labels, using n^2 algo!!! */
      for (stmt = firstStatement; stmt; stmt = stmt->nextLink) {
	if ((stmt->kind == PDB_EXIT || stmt->kind == PDB_CYCLE) && stmt->code->ext.whichloop) {
	  for (tmpstmt = firstStatement; tmpstmt; tmpstmt = tmpstmt->nextLink) {

	    /* This is stupid, why is there a duplicate gfc_code*?  The exit/cycle points to a different one
	       than the DO loop actually has.  For now, I'm just going to check the linenumber.  I'm pretty
	       sure you can only have one label/DO loop on a line. */

/* 	    if (tmpstmt->kind == PDB_DO && stmt->code->ext.whichloop == tmpstmt->code) { */
	    if (tmpstmt->kind == PDB_DO && stmt->code->ext.whichloop->loc.pdbStartLine == tmpstmt->code->loc.pdbStartLine) {
	      stmt->extrastmt = tmpstmt;
	    }
	  }
	}
      }





/*       fprintf (pdbfile, "rstmt st#%d freturn so#%d %d %d NULL 0 0 st#%d NA\n", tau_st_num, tau_so_num, end.tau_line, theColumn, tau_st_num+1); */
/*       tau_st_num++; */



      rstart = ns->tau_progunit->tau_rstart_locus;
      theColumn = rstart.pdbStartColumn;
      if (rstart.lb) {
	theColumn -= rstart.lb->tab;
      }
      routine->location.file = tau_so_num;
      routine->location.line = rstart.pdbStartLine;
      routine->location.column = theColumn;


      locus rpos_start = ns->tau_progunit->tau_rpos_start_locus;
      theColumn = rpos_start.pdbStartColumn;
      if (rpos_start.lb) {
	theColumn -= rpos_start.lb->tab;
      }
      routine->rpos_start.file = tau_so_num;
      routine->rpos_start.line = rpos_start.pdbStartLine;
      routine->rpos_start.column = theColumn;


      locus rpos_end = ns->tau_progunit->tau_rpos_end_locus;
      theColumn = rpos_end.pdbEndColumn;
      if (rpos_end.lb) {
	theColumn -= rpos_end.lb->tab;
      }
      routine->rpos_end.file = tau_so_num;
      routine->rpos_end.line = rpos_end.pdbEndLine;
      routine->rpos_end.column = theColumn;


      /* temp hack for empty routines */
      if (routine->location.line < 1) {

	end = ns->tau_progunit->tau_end_locus; 
	theColumn = end.pdbStartColumn;
	if (end.lb->tab) {
	  theColumn -= 6;
	}

	routine->location.line = end.pdbEndLine;
	routine->location.column = theColumn;
      }

      routine->statements = firstStatement;
      routine->rcalls = pdbFirstRcall;
      routine->appearsInThisFile = 1;

#ifdef PDB_ENABLE_TYPES
      /* Warning! recursive derived types cause infinite loop */
      routine->type = createType(routine->sym);
#endif

    } else {

/*       printf ("hmm\n"); */
    }

  } else {
/*     printf ("\nhmm2\n"); */
  }

  /* recursively enter contained namespaces */
  for (ns = ns->contained; ns; ns = ns->sibling) {
    pdt_output_pdb(ns);
  }

/*   printf("\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-\n\n"); */


}


void outputType(pdbType *type);
void outputType(pdbType *type) {
  pdbArg *arg;
  if (!type) return;

  fprintf (pdbfile, "%s %s\n", type->idString, type->name);

  if (type->kind != KIND_DERIVED) {
    fprintf (pdbfile, "yacs priv\n");
    fprintf (pdbfile, "ykind %s\n", ykind_strings[type->kind]);
  }

  switch (type->kind) {
  case KIND_INT:
    fprintf (pdbfile, "yikind %s\n", yikind_strings[type->ikind]);
    break;
  case KIND_FLOAT:
    fprintf (pdbfile, "yfkind %s\n", yfkind_strings[type->fkind]);
    break;
  case KIND_DERIVED:
    
    fprintf (pdbfile, "gkind %s\n", ykind_strings[type->kind]);
    for (arg = type->args; arg; arg=arg->nextLink) {
      fprintf (pdbfile, "gmem %s\n", arg->name);
      /* gmloc? */
      fprintf (pdbfile, "gmacs %s\n", access_strings[arg->type->access]);
      fprintf (pdbfile, "gmkind var\n"); /* ? */
      fprintf (pdbfile, "gmtype %s\n", arg->type->idString);
    }
    break;
    
  case KIND_FUNC:
    fprintf (pdbfile, "yrett %s\n", type->returnType->idString);
    for (arg = type->args; arg; arg=arg->nextLink) {
      fprintf (pdbfile, "yargt %s %s\n", arg->type->idString, arg->name);
    }
    break;

  case KIND_ARRAY:
    fprintf (pdbfile, "yelem %s\n", type->baseType->idString);
    fprintf (pdbfile, "yshape %s\n", yshape[type->as.type]);
    fprintf (pdbfile, "yrank %d\n", type->as.rank);
    break;

  case KIND_POINTER:
    fprintf (pdbfile, "yptr %s\n", type->baseType->idString);
    break;

  case KIND_VOID:
    break;
  case KIND_NA:
  case KIND_LOGICAL:
    fprintf (pdbfile, "yikind %s\n", yikind_strings[type->ikind]);
  case KIND_CHAR:
  case KIND_COMPLEX:
  case KIND_UNSPECFUNC:
  case KIND_BLDAT:
  case KIND_MOD:
    fprintf (stderr, "gfortran: warning: unsupported type: %s\n", type->name);
  }


  fprintf (pdbfile, "\n");
}

void pdt_write_pdb();
void pdt_write_pdb() {
  pdbRoutine *routine;
  pdbRcall *rcall;
  pdbStatement *stmt;
  pdbType *type;

/*   printf ("Writing PDB file\n"); */
/*   printf ("=============== pdt_write_pdb =================\n"); */
  fprintf (pdbfile, "\n");
  for (routine = pdbFirstRoutine; routine; routine = routine->nextLink) {
/*     printf ("ROUTINE: %s\n", routine->name); */
    
    if (!routine->appearsInThisFile) {
      continue;
    }


    fprintf (pdbfile, "ro#%d %s\n", routine->id, routine->name);

#ifdef PDB_ENABLE_TYPES
    if (routine->type) {
      fprintf (pdbfile, "rsig %s\n", routine->type->idString);
    }
#endif    

    if (routine->appearsInThisFile) {
      fprintf (pdbfile, "rstart so#%d %d %d\n", routine->location.file, 
	       routine->location.line, routine->location.column);
/*       fprintf (pdbfile, "rloc so#%d %d %d\n",  routine->location.file,  */
/* 	       routine->location.line, routine->location.column); */
      fprintf (pdbfile, "rloc so#%d %d %d\n",  routine->rpos_start.file, 
	       routine->rpos_start.line, routine->rpos_start.column);

      fprintf (pdbfile, "rpos so#%d %d %d so#%d %d %d so#%d %d %d so#%d %d %d\n",  
	       routine->rpos_start.file, 
	       routine->rpos_start.line, 
	       routine->rpos_start.column,
	       routine->rpos_start.file, 
	       routine->rpos_start.line, 
	       routine->rpos_start.column,

	       routine->rpos_end.file, 
	       routine->rpos_end.line, 
	       routine->rpos_end.column,
	       routine->rpos_end.file, 
	       routine->rpos_end.line, 
	       routine->rpos_end.column
	       );

    }
    fprintf (pdbfile, "rlink f90\n");
    if (routine->sym->attr.flavor == FL_PROGRAM) {
      fprintf (pdbfile, "rkind fprog\n");
    } else {
      fprintf (pdbfile, "rkind fext\n");
    }
    
    if (gfc_elemental(routine->sym)) {
      fprintf (pdbfile, "rfprefix elem\n");
    } else if (gfc_pure(routine->sym)) {
      fprintf (pdbfile, "rfprefix pure\n");
    }
    
#ifdef PDB_ENABLE_RCALL
    for (rcall = routine->rcalls; rcall; rcall = rcall->nextLink) {
      outputRcall(rcall);
    }
#endif
    
    
    for (stmt = routine->statements; stmt; stmt = stmt->nextLink) {
	outputStatement(stmt);
    }
    
    
    if (routine->statements) {
      fprintf (pdbfile, "rbody st#0\n");
    }
    fprintf (pdbfile, "\n");
  }


#ifdef PDB_ENABLE_TYPES
  // Type information
  for (type = pdbFirstType; type; type = type->nextLink) {
    outputType(type);
  }
#endif

}

